from collections import defaultdict
from ipaddress import ip_network
from typing import Any

from pydantic import BaseModel
from pytricia import PyTricia

from ipfabric_infoblox.config.network_view import NetworkView
from ipfabric_infoblox.ipf_models import ManagedNetwork, ManagedIP, Export
from ipfabric_infoblox.sync.sync import Sync
from .models import Log


class NetworkValidation(BaseModel):
    view_config: NetworkView
    network_view: str
    networks_pyt: Any
    containers_pyt: Any
    sync: Sync

    def _copy_pyt(self, network) -> PyTricia:
        if self.networks_pyt.has_key(network):
            return self.networks_pyt
        else:
            tmp = PyTricia()
            [tmp.insert(_, self.networks_pyt[_]) for _ in self.networks_pyt]
            tmp.insert(network, network)
            return tmp

    def _validate_network(self, log: Log, policy: str) -> Log:
        network = log.network
        ip = ip_network(network)
        if ip.version == 4 and ip.prefixlen > log.smallest_v4_subnet:
            log.skip_reason = f"Skipping IPF subnet '{network}' because mask is greater than `smallest_v4_subnet={log.smallest_v4_subnet}`."
        elif self._copy_pyt(network).children(network):
            log.has_child_network = True
            log.failure = f"Failed for IPF subnet '{network}' because IB has children network(s)."
        elif network in self.networks_pyt and self.networks_pyt[network].has_hosts:
            log.parent_network = self.networks_pyt[network]
            log.failure = f"Failed for IPF subnet '{network}' because parent IB Network '{self.networks_pyt[network].network}' has hosts."
        elif network in self.networks_pyt and not log.split_networks:
            log.parent_network = self.networks_pyt[network]
            log.skip_reason = f"Skipping IPF subnet '{network}' under parent IB Network '{self.networks_pyt[network].network}' because `{policy}` is True and `split_networks` is False."
        elif network in self.networks_pyt:
            log.parent_network = self.networks_pyt[network]
        return log

    def validate_network(self, ipf_network: ManagedNetwork) -> Log:
        network = str(ipf_network.network)
        log = Log(
            network=network,
            network_view=self.network_view,
            create_containerless_nets=self.view_config.create_containerless_nets,
            split_networks=self.view_config.split_networks,
            smallest_v4_subnet=self.view_config.smallest_v4_subnet,
            ipf_networks={ipf_network},
        )
        if self.containers_pyt.has_key(network):
            log.network_container = self.containers_pyt[network]
            log.failure = (
                f"Failed for IPF subnet '{network}' because it is an IB Container '{log.network_container.network}'."
            )
        elif network in self.containers_pyt and self.networks_pyt.has_key(network):
            log.network_container = self.containers_pyt[network]
            log.network_match = self.networks_pyt[network]
        elif network in self.containers_pyt:
            log.network_container = self.containers_pyt[network]
            log = self._validate_network(log, "ib_create_network")
        elif self.networks_pyt.has_key(network):
            log.network_match = self.networks_pyt[network]
        elif not self.view_config.create_containerless_nets:
            log.skip_reason = f"Skipping IPF subnet '{network}' because `create_containerless_nets` is False."
        else:
            log = self._validate_network(log, "ib_create_container")

        self.sync.logs.append(log)
        return log


class FinalValidation(BaseModel):
    @staticmethod
    def _validate_logs(logs: set[Log]) -> set[Log]:
        pyt = PyTricia()
        skip, create = set(), set()
        for log in logs:
            if log.create:
                pyt[log.network] = log
            else:
                skip.add(log)

        for net in pyt:
            if pyt.parent(net) or pyt.children(net):
                pyt[net].failure = f"Failed for IPF subnet '{net}' because overlapping IP Fabric networks found."
            create.add(pyt[net])
        return skip | create

    def validate_logs(self, logs: list[Log], pyt_validate: bool = True) -> list[Log]:
        views = defaultdict(dict)
        for log in logs:
            if str(log) in views[log.network_view]:
                views[log.network_view][str(log)].ipf_networks.update(log.ipf_networks)
            else:
                views[log.network_view][str(log)] = log
        if not pyt_validate:
            return [log for logs in views.values() for log in logs.values()]

        final = list()
        for view, logs in views.items():
            tmp = {views[view][_] for _ in logs}
            final.extend(list(self._validate_logs(tmp)))
        return final

    def validate_ips(self, managed_ips: dict[str, list[ManagedIP]]) -> dict[str, list[Export]]:
        validated = {_: list() for _ in managed_ips}
        for view, ips in managed_ips.items():
            tmp = defaultdict(list)
            for ip in ips:
                tmp[ip.ip].append(ip)
            for ip, mips in tmp.items():
                if len(mips) == 1:
                    validated[view].append(mips[0].export())
                else:
                    validated[view].append(ManagedIP.join_ips(mips))
        return validated



from typing import Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric_infoblox.ipf_models import ManagedNetwork


class Ref(BaseModel):
    ref: str = Field(alias="_ref")
    comment: Optional[str] = None


class View(Ref, BaseModel):
    name: str
    is_default: bool


class Container(Ref, BaseModel):
    network: str
    network_view: str


class Network(Container, BaseModel):
    network_container: str
    conflict_count: int
    dynamic_hosts: int
    static_hosts: int
    total_hosts: int
    unmanaged_count: int

    @property
    def has_hosts(self) -> bool:
        return (
            self.total_hosts + self.conflict_count + self.dynamic_hosts + self.static_hosts + self.unmanaged_count
        ) > 0


class Log(BaseModel):
    network: str
    network_view: str
    create_containerless_nets: bool = False
    split_networks: bool = False
    smallest_v4_subnet: int = 31
    ipf_networks: set[ManagedNetwork]
    network_container: Optional[Container] = None
    parent_network: Optional[Network] = None
    network_match: Optional[Network] = None
    created_network: Optional[Network] = None
    has_child_network: bool = False
    failure: Optional[str] = None
    skip_reason: Optional[str] = None

    def __str__(self):
        return f"{self.network} - {self.network_view}"

    def __hash__(self):
        return hash(str(self))

    @computed_field
    @property
    def has_parent_network(self) -> bool:
        return self.parent_network is not None

    @computed_field
    @property
    def parent_network_has_hosts(self) -> bool:
        return self.parent_network.has_hosts if self.parent_network else False

    @computed_field
    @property
    def network_is_container(self) -> bool:
        return self.network_container.network == self.network if self.network_container else False

    @computed_field
    @property
    def has_network_match(self) -> bool:
        return self.network_match is not None

    @computed_field
    @property
    def skipped(self) -> bool:
        return self.skip_reason is not None

    @computed_field
    @property
    def failed(self) -> bool:
        return self.failure is not None

    @computed_field
    @property
    def create(self) -> bool:
        return not self.failed and not self.skipped and not self.has_network_match

    @computed_field
    @property
    def status(self) -> str:
        """Dynamically computed status based on log properties."""
        if self.failed:
            return "failed"
        elif self.skipped:
            return "skipped"
        elif self.create:
            return "create"
        elif self.has_network_match:
            return "exists"
        return "unknown"



"""https://www.dvolve.net/blog/2020/09/hacking-infoblox-discovery-for-fun-and-profit/"""

from ipaddress import IPv4Interface
from typing import Optional

from ipfabric.models import Device, Snapshot
from macaddress import MAC
from pydantic import BaseModel, ConfigDict, field_validator

from ipfabric_infoblox.config import DEFAULT_VRF


def convert_none(value: Optional[str]) -> str:
    return value or ""


class VLAN(BaseModel):
    sn: str
    vlanId: int
    vlanName: Optional[str] = ""
    model_config = ConfigDict(extra="ignore")
    _normalize_vlan = field_validator("vlanName")(convert_none)


class VRF(BaseModel):
    sn: str
    vrf: str
    intName: str
    rd: Optional[str] = ""
    model_config = ConfigDict(extra="ignore")
    _normalize_rd = field_validator("rd")(convert_none)


class SNMP(BaseModel):
    sn: str
    location: Optional[str] = ""
    contact: Optional[str] = ""
    model_config = ConfigDict(extra="ignore")
    _normalize_snmp = field_validator("location", "contact")(convert_none)


class Interface(BaseModel):
    sn: str
    intName: str
    nameOriginal: Optional[str] = ""
    dscr: Optional[str] = ""
    duplex: Optional[str] = ""
    speed: Optional[str] = ""
    media: Optional[str] = ""
    model_config = ConfigDict(extra="ignore")
    _normalize_int = field_validator("dscr", "duplex", "media")(convert_none)

    @property
    def calc_duplex(self) -> str:
        # Bad syntax for member 'port_duplex' in 'discovery_data' val:'auto' on insert err:'Values must be 'Full Half''
        if self.duplex in ["full", "half"]:
            return self.duplex.capitalize()
        return ""


class Export(BaseModel):
    ip_address: str
    device_contact: str = ""
    device_location: str = ""
    device_model: str = ""
    device_vendor: str = ""
    device_port_name: str = ""
    device_port_type: str = ""
    discovered_name: str = ""
    discoverer: str = "IP Fabric"
    last_discovered_timestamp: str = ""
    mac_address: str = ""
    network_component_model: str = ""
    network_component_name: str = ""
    network_component_port_description: str = ""
    network_component_port_name: str = ""
    network_component_type: str = ""
    network_component_vendor: str = ""
    os: str = ""
    port_duplex: str = ""
    port_link_status: str = ""
    port_status: str = ""
    port_vlan_name: str = ""
    port_vlan_number: str = ""
    vrf_name: str = ""
    vrf_rd: str = ""
    task_name: str = "IP Fabric Snapshot"
    # ap_ip_address="",  # TODO: WIFI ARP Stuff
    # ap_name="",  # TODO: WIFI ARP Stuff
    # ap_ssid="",  # TODO: WIFI ARP Stuff
    # bgp_as="",  # TODO: Have to combine BGP advertised routes and neighbors?
    # bridge_domain="",  # TODO: ACI ARP Stuff
    # tenant="",  # TODO: ACI ARP Stuff
    # endpoint_groups="",  # TODO: ACI ARP Stuff


class ManagedIP(BaseModel):
    hostname: str
    ip: str
    sn: str
    intName: str
    stateL1: Optional[str] = ""
    stateL2: Optional[str] = ""
    siteName: str
    vlanId: Optional[int] = None
    net: IPv4Interface
    mac: Optional[str] = ""
    device: Device
    vrf: str = ""
    type: Optional[str] = ""
    snmp: Optional[SNMP] = None
    interface: Optional[Interface] = None
    snapshot: Snapshot
    vrf_obj: Optional[VRF] = None
    vlan_obj: Optional[VLAN] = None
    _normalize_ip = field_validator("stateL1", "stateL2", "mac", "vrf", "type")(convert_none)
    model_config = ConfigDict(extra="ignore")

    @property
    def calc_vrf(self) -> str:
        if self.device.vendor not in DEFAULT_VRF:
            default_vrf = ""
        elif self.device.family not in DEFAULT_VRF[self.device.vendor]:
            default_vrf = DEFAULT_VRF[self.device.vendor].get(None, "")
        else:
            default_vrf = DEFAULT_VRF[self.device.vendor][self.device.family]
        return self.vrf if self.vrf != default_vrf else ""

    @staticmethod
    def calc_state(state) -> str:
        # Bad syntax for member 'port_link_status':'Values must be 'Connected Not\ Connected Unknown''
        if state in ["up"]:
            return "Connected"
        if state in ["down"]:
            return "Not Connected"
        return "Unknown"

    def export(self) -> Export:
        # wapidoc/additional/structs.html#struct-discoverydata
        return Export(
            ip_address=self.ip,
            device_contact=self.snmp.contact if self.snmp else "",
            device_location=self.snmp.location if self.snmp else "",
            device_model=self.device.model or "",
            device_vendor=self.device.vendor,
            device_port_type=self.interface.media if self.interface else self.intName,
            discovered_name=self.device.fqdn if self.device.fqdn else self.hostname,
            last_discovered_timestamp=(
                self.device.ts_discovery_start if self.device.ts_discovery_start else self.snapshot.start
            ).strftime("%Y-%m-%d %H:%M:%S"),
            mac_address=str(MAC(self.mac)).replace("-", ":").lower() if self.mac else "",
            network_component_model=self.device.model or "",
            network_component_name=self.hostname,
            network_component_port_description=self.interface.dscr if self.interface else "",
            network_component_port_name=self.intName,
            network_component_type=self.device.dev_type,
            network_component_vendor=self.device.vendor,
            os=self.device.version,
            port_duplex=self.interface.calc_duplex if self.interface else "",
            port_link_status=self.calc_state(self.stateL1),
            port_status=self.stateL2,
            port_vlan_name=self.vlan_obj.vlanName if self.vlan_obj else "",
            port_vlan_number=str(self.vlanId or ""),
            vrf_name=self.calc_vrf,
            vrf_rd=self.vrf_obj.rd if self.vrf_obj else "",
            task_name=f"IP Fabric Snapshot {self.snapshot.snapshot_id}",
        )

    @classmethod
    def join_ips(cls, ips: list["ManagedIP"]) -> Export:
        joined = dict(
            port_vlan_name=set(),
            port_vlan_number=set(),
            vrf_name=set(),
            vrf_rd=set(),
            mac_address=set(),
        )
        for ip in ips:
            for k, v in ip.export().model_dump().items():
                if k in joined and v:
                    joined[k].add(v)
        for k, v in joined.items():
            joined[k] = v.pop() if len(v) == 1 else ""

        ip = ips[0]
        return Export(
            ip_address=ip.ip,
            discovered_name=",".join({_.device.fqdn if _.device.fqdn else _.hostname for _ in ips}),
            last_discovered_timestamp=ip.snapshot.start.strftime("%Y-%m-%d %H:%M:%S"),
            network_component_name=",".join({_.hostname for _ in ips}),
            task_name=f"IP Fabric Snapshot {ip.snapshot.snapshot_id}",
            **joined,
        )


"""
Do not use the following:
Bad syntax for member 'network_component_port_number' in 'discovery_data' val:'0/0/2.0' on insert err:'Invalid integer.'
network_component_port_number=self.interface.calc_name[1]

Bad syntax for member 'port_speed' in 'discovery_data' val:'auto' on insert err:'Values must be '10M 100M 1G 10G 100G Unknown''
port_speed=self.interface.calc_speed,
"""



from typing import Any, Optional

from pydantic import BaseModel, Field, PrivateAttr, model_validator
from pytricia import PyTricia

from .infoblox import Infoblox
from .ipfabric import IPFabric
from .network_view import NetworkView


class Configuration(BaseModel):
    ipfabric: Optional[IPFabric] = Field(default_factory=IPFabric)
    infoblox: Optional[Infoblox] = Field(default_factory=Infoblox)
    network_views: list[NetworkView] = Field(alias="networkViews")
    dry_run: bool = True
    _pyt: Any = PrivateAttr(default_factory=PyTricia)

    def model_post_init(self, __context: Any) -> None:
        self.build_pyt()

    @model_validator(mode="before")
    @classmethod
    def check_view_names(cls, data: Any) -> Any:
        if isinstance(data, dict) and "networkViews" in data:
            view_names = [_["name"] for _ in data["networkViews"]]
            assert len(view_names) == len(set(view_names)), "Duplicate View Names found."
        return data

    @model_validator(mode="after")
    def check_config(self):
        if len(self.network_views) == 0:
            raise ValueError("No Network Views configured.")
        if len([_ for _ in self.network_views if _.default]) > 1:
            raise ValueError(
                f"Only 1 default Network View can be configured and multiple found {list(self.default_view)}."
            )
        return self

    @property
    def default_view(self) -> NetworkView:
        return [_ for _ in self.network_views if _.default][0]

    @property
    def view_dict(self) -> dict[str, NetworkView]:
        return {_.name: _ for _ in self.network_views}

    @property
    def pyt(self):
        return self._pyt

    def build_pyt(self):
        for view in self.network_views:
            for network in view.include.networks:
                if self._pyt.has_key(network):
                    print(f"Duplicate network: {network}")
                    self._pyt[network].append(view.name)
                else:
                    self._pyt[network] = [view.name]



import csv
from pathlib import Path
from typing import Any

from ipfabric import IPFClient
from pydantic import BaseModel, PrivateAttr

from ipfabric_infoblox.config import Configuration
from ipfabric_infoblox.infoblox.models import Log, Field
from ipfabric_infoblox.ipf_models.managed_ip import ManagedIP
from ipfabric_infoblox.sync.ip_sync import IPSync
from ipfabric_infoblox.sync.network_sync import NetworkSync


class Sync(BaseModel):
    config: Configuration
    logs: list[Log] = Field(default_factory=list)
    _ipf: IPFClient = PrivateAttr(None)
    _network_sync: NetworkSync = PrivateAttr(None)
    _ip_sync: list[ManagedIP] = PrivateAttr(default_factory=list)

    def model_post_init(self, __context: Any) -> None:
        self._ipf = IPFClient(
            base_url=self.config.ipfabric.base_url,
            auth=(
                self.config.ipfabric.token
                if self.config.ipfabric.token
                else (self.config.ipfabric.username, self.config.ipfabric.password)
            ),
            verify=self.config.ipfabric.verify,
            timeout=self.config.ipfabric.timeout,
            snapshot_id=self.config.ipfabric.snapshot_id,
        )
        self._network_sync = NetworkSync(ipf=self._ipf, config=self.config)
        self._ip_sync = IPSync(ipf=self._ipf).build()

    @property
    def network_sync(self) -> NetworkSync:
        return self._network_sync

    @property
    def ip_sync(self) -> list[ManagedIP]:
        return self._ip_sync

    def export_logs_to_csv(self, file_path: str) -> None:
        file_path = Path(file_path)
        with file_path.open(mode="w", newline="") as csv_file:
            fieldnames = list(Log.model_json_schema()["properties"].keys())
            computed_fields = [
                "parent_network_has_hosts",
                "has_parent_network",
                "has_network_match",
                "network_is_container",
                "failed",
                "status",
                "skipped",
                "create",
            ]
            fieldnames.extend(computed_fields)
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()
            for log in self.logs:
                log_dict = log.model_dump()
                writer.writerow(log_dict)



from ipaddress import IPv4Network
from typing import Optional, Any, Literal, Union

from pydantic import BaseModel, Field, model_validator, field_validator, PrivateAttr, field_serializer
from pytricia import PyTricia

DEFAULT_NET = "0.0.0.0/0"
IP_LITERALS = Literal["RFC1918", "CGNAT"]
IP_TYPES = Optional[Union[IPv4Network, IP_LITERALS, list[Union[IPv4Network, IP_LITERALS]]]]
IP_LITERAL = {
    "CGNAT": {IPv4Network("100.64.0.0/10")},
    "RFC1918": {IPv4Network("10.0.0.0/8"), IPv4Network("172.16.0.0/12"), IPv4Network("192.168.0.0/16")},
}


def format_networks(data: Any) -> list[Union[str, IPv4Network]]:
    new_data = set()
    for _ in [data] if isinstance(data, str) else data:
        if _ and isinstance(_, str) and _.upper() in IP_LITERAL:
            new_data.update(IP_LITERAL[_.upper()])
        elif _:
            new_data.add(_)
    return list(new_data)


class StringConfig(BaseModel):
    value: str
    regex: bool = False
    ignore_case: bool = True


class ConfigList(BaseModel):
    vrfs: Optional[list[StringConfig]] = Field(default_factory=list)
    sites: Optional[list[StringConfig]] = Field(default_factory=list)
    networks: IP_TYPES = Field(default_factory=list)

    @field_validator("networks", mode="before")
    @classmethod
    def check_networks(cls, data: Any) -> list[Union[str, IPv4Network]]:
        return format_networks(data)

    @field_serializer("networks", when_used="json")
    def serialize_networks(self, networks: list[Any]):
        return [str(_) for _ in networks]

    @property
    def has_items(self) -> bool:
        return True if self.networks or self.vrfs or self.sites else False

    @property
    def vrf_names(self) -> list[StringConfig]:
        return [_ for _ in self.vrfs if not _.regex]

    @property
    def vrf_regex(self) -> list[StringConfig]:
        return [_ for _ in self.vrfs if _.regex]

    @property
    def site_names(self) -> list[StringConfig]:
        return [_ for _ in self.sites if not _.regex]

    @property
    def site_regex(self) -> list[StringConfig]:
        return [_ for _ in self.sites if _.regex]


class Include(ConfigList, BaseModel): ...


class Exclude(ConfigList, BaseModel): ...


class NetworkView(BaseModel):
    name: str
    default: bool = False
    include: Optional[Include] = Field(default_factory=Include)
    exclude: Optional[Exclude] = Field(default_factory=Exclude)
    create_containerless_nets: bool = False
    managed_ip_discovery: bool = True
    split_networks: bool = False
    smallest_v4_subnet: int = 31
    _pyt: PyTricia = PrivateAttr(default_factory=PyTricia)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        return self.name == other.name

    def model_post_init(self, __context: Any) -> None:
        if self.default or not self.include.networks:
            self._pyt.insert(DEFAULT_NET, "INCLUDE")
        else:
            for _ in self.include.networks:
                self._pyt.insert(_, "INCLUDE")
        for _ in self.exclude.networks:
            self._pyt.insert(_, "EXCLUDE")

    @model_validator(mode="after")
    def check_config(self):
        if self.default and self.include.has_items:
            raise ValueError("Default VRF can only exclude Networks, VRFs, or Site Names and cannot include items.")
        if not self.default and not self.include.has_items:
            raise ValueError("Non-default VRF must include at least one Network, VRF, or Site Name.")
        return self

    @property
    def include_has_vrf(self) -> bool:
        return True if self.include.vrfs else False

    @property
    def exclude_has_vrf(self) -> bool:
        return True if self.exclude.vrfs else False

    @property
    def include_has_sites(self) -> bool:
        return True if self.include.sites else False

    @property
    def exclude_has_sites(self) -> bool:
        return True if self.exclude.sites else False

    @property
    def pyt(self):
        return self._pyt



import os
from collections import defaultdict
from typing import Annotated

import typer
import yaml
from rich.console import Console

from ipfabric_infoblox.config import Configuration
from ipfabric_infoblox.infoblox.ib_validate import NetworkValidation, FinalValidation
from ipfabric_infoblox.infoblox.infoblox import Infoblox
from ipfabric_infoblox.sync.sync import Sync

app = typer.Typer()
console = Console()

DEFAULT_LOG_FILE = "diff.csv"


def print_error_and_abort(message: str):
    console.print(f"Error: {message}", style="bold red")
    raise typer.Abort()


def load_yaml_config(config_file: typer.FileText) -> dict:
    try:
        config_data = yaml.safe_load(config_file)
        return config_data
    except yaml.YAMLError as exc:
        typer.echo(f"Error parsing YAML file: {exc}", err=True)
        raise typer.Exit(code=1)


def common_logic(config_file: typer.FileText, logging: bool):
    console.print(
        f"Comparing networks in [link={os.environ.get('IPF_URL')}]IP Fabric[/link] and [link={os.environ.get('IB_HOST')}]Infoblox[/link]",
        style="bold green",
    )
    config_data = load_yaml_config(config_file)
    config = Configuration(**config_data)
    sync = Sync(config=config)
    ib = Infoblox(config=config)
    for view in config.network_views:
        if view.name not in ib.views:
            raise ValueError(f"View {view.name} does not exists in IB.")
    views = {
        _.name: NetworkValidation(
            view_config=_,
            network_view=_.name,
            networks_pyt=ib.networks_pyt(_.name),
            containers_pyt=ib.containers_pyt(_.name),
            sync=sync,
        )
        for _ in config.network_views
    }
    validated_networks, matched_networks, ib_csv, ib_ip_csvs = [], [], None, None
    for network in sync.network_sync.networks.networks:
        if not network.success.network_view:
            console.print(f"Error: Network {network.network} failed to validate.", style="bold red")
            continue
        validated = views[network._net_view].validate_network(network)
        if validated.create:
            validated_networks.append(validated)
        if validated.has_network_match:
            matched_networks.append(validated)

    validated_networks = FinalValidation().validate_logs(
        logs=validated_networks
    )  # TODO: This fixes duplicate networks but now all the logging is screwed up
    matched_networks = FinalValidation().validate_logs(logs=matched_networks, pyt_validate=False)

    for log in sync.logs:
        console.print(f"    {log.network} - {log.status} {log.skip_reason if log.skip_reason else ''}")

    if logging:
        sync.export_logs_to_csv(DEFAULT_LOG_FILE)
        console.print(f"Logs exported to {DEFAULT_LOG_FILE}", style="bold green")

    create_nets = [_ for _ in validated_networks if _.create]
    if create_nets:
        console.print("Creating Infoblox Network CSV import CSV file.", style="bold green")
        ib_csv = ib.create_csv(logs=create_nets, comment="Synced from IP Fabric")

    ipf_nets = dict()
    for log in create_nets + matched_networks:
        for net in log.ipf_networks:
            site = ipf_nets.setdefault(net.site_name, dict())
            vrf = site.setdefault(net.vrf, dict())
            vrf[str(net.network)] = log

    validated_ip = defaultdict(list)
    for ip in sync.ip_sync:
        # TODO: What to do with /31 and /32
        log = ipf_nets.get(ip.siteName, {}).get(ip.vrf, {}).get(str(ip.net.network))
        if log:
            validated_ip[log.network_view].append(ip)
    validated_ip = FinalValidation().validate_ips(validated_ip)

    if validated_ip:
        console.print("Creating Infoblox IP Discovery CSV import CSV file.", style="bold green")
        ib_ip_csvs = {
            k: ib.create_ip_csv(v) for k, v in validated_ip.items() if views[k].view_config.managed_ip_discovery
        }

    return config, sync, ib, ib_csv, ib_ip_csvs


@app.command(name="diff")
def diff_cmd(
    config: Annotated[
        typer.FileText, typer.Option(help="YAML file with configuration default = config.yml")
    ] = "config.yml",
    logging: Annotated[bool, typer.Option(help="Say hi formally.")] = False,
):
    """Diff command to compare configurations."""
    common_logic(config, logging)


@app.command(name="sync")
def sync_cmd(
    config: Annotated[
        typer.FileText, typer.Option(help="YAML file with configuration default = config.yml")
    ] = "config.yml",
    logging: Annotated[bool, typer.Option(help="Say hi formally.")] = False,
):
    """Sync command to compare configurations."""
    config, _, ib, ib_csv, ib_ip_csvs = common_logic(config, logging)

    if not ib_csv:
        console.print("No networks to sync", style="bold red")
    else:
        data = ib.csv_upload(ib_csv)
        # TODO: Error checking of upload
    if not ib_ip_csvs:
        console.print("No IPs to sync", style="bold red")
    else:
        for view, ips in ib_ip_csvs.items():
            ib.csv_discovery_upload(ips, view)


if __name__ == "__main__":
    app()



import re
from typing import Any

from pydantic import BaseModel, PrivateAttr, Field

from ipfabric_infoblox.config import Configuration, DEFAULT_VRF
from ipfabric_infoblox.config.network_view import NetworkView
from ipfabric_infoblox.ipf_models.managed_networks import ManagedNetworks, ManagedNetwork


class Map(BaseModel):
    include: set[NetworkView] = Field(default_factory=set)
    exclude: set[NetworkView] = Field(default_factory=set)


class NetworkSync(BaseModel):
    ipf: Any
    config: Configuration
    _networks: ManagedNetworks = PrivateAttr(None)
    _site_map: dict[str, Map] = PrivateAttr(default_factory=dict)
    _vrf_map: dict[str, Map] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any) -> None:
        self._networks = self._get_ipv4_nets()
        self._site_map = {_: Map(**self._mapping(_, "sites")) for _ in self._networks.sites}
        self._vrf_map = {_: Map(**self._mapping(_, "vrfs")) for _ in self._networks.vrfs}
        self.map_net_view()
        self.map_vrf_view()
        self.map_site_view()
        self.map_default_view()

    @property
    def networks(self) -> ManagedNetworks:
        return self._networks

    @staticmethod
    def _calculate_vrf(vrf: str, dev: dict) -> str:
        if dev["vendor"] not in DEFAULT_VRF:
            default_vrf = ""
        elif dev["family"] not in DEFAULT_VRF[dev["vendor"]]:
            default_vrf = DEFAULT_VRF[dev["vendor"]].get(None, "")
        else:
            default_vrf = DEFAULT_VRF[dev["vendor"]][dev["family"]]
        return vrf if vrf != default_vrf else ""

    def _get_ipv4_nets(self) -> ManagedNetworks:
        devs = {_["sn"]: _ for _ in self.ipf.inventory.devices.all(columns=["sn", "vendor", "family"])}

        networks = set()
        vrf_map = self.config.ipfabric.mapped_vrfs
        for net in self.ipf.technology.addressing.managed_ip_ipv4.all(
            columns=["siteName", "net", "vrf", "sn"], filters={"net": ["empty", False]}
        ):
            if devs[net["sn"]]["vendor"] in ["aws", "azure", "gcp"]:
                continue  # TODO: Do we want to exclude Cloud?
            net["vrf"] = self._calculate_vrf(net["vrf"], devs[net["sn"]])
            net["mapped_vrf"] = net["vrf"]
            if net["vrf"] in vrf_map or net["vrf"].lower() in vrf_map:
                net["mapped_vrf"] = vrf_map[net["vrf"]] if net["vrf"] in vrf_map else vrf_map[net["vrf"].lower()]

            networks.add(ManagedNetwork(**net))
        return ManagedNetworks(networks=list(networks))

    @staticmethod
    def _map(value, test) -> bool:
        if (
            (test.regex and re.match(test.value, value, flags=re.IGNORECASE if test.ignore_case else 0))
            or (test.ignore_case and test.value.lower() == value.lower())
            or test.value == value
        ):
            return True
        return False

    def _mapping(self, value: str, attr: str) -> dict[str, set[NetworkView]]:
        include, exclude = set(), set()
        for view in self.config.network_views:
            for inc in getattr(view.include, attr):
                if self._map(value, inc):
                    include.add(view)
            for exc in getattr(view.exclude, attr):
                if self._map(value, exc):
                    exclude.add(view)
        return {"include": include - exclude, "exclude": exclude}

    def _is_site_included(self, site: str, view: str) -> bool:
        if view not in self._site_map[site].exclude and (
            (self._site_map[site].include and view in self._site_map[site].include) or not self._site_map[site].include
        ):
            return True
        return False

    def _is_vrf_included(self, vrf: str, view: str) -> bool:
        if (
            vrf in self._vrf_map
            and view not in self._vrf_map[vrf].exclude
            and ((self._vrf_map[vrf].include and view in self._vrf_map[vrf].include) or not self._vrf_map[vrf].include)
        ):
            return True
        return False

    @staticmethod
    def _is_net_allowed(net, view: NetworkView) -> bool:
        return True if net in view.pyt and view.pyt[net] == "INCLUDE" else False

    @staticmethod
    def _check_views(network: ManagedNetwork, tmp: set[str], error_gt_1: str, error_eq_0: str, success: str):
        if len(tmp) == 1:
            network.net_view = tmp.pop()
            network.error.multiple_views, network.error.no_matching_view = False, False
            network.success.network_view = success
        elif len(tmp) > 1:
            network.error.multiple_views = True
            network.error.message.append(error_gt_1)
        elif not tmp:
            network.error.no_matching_view = True
            network.error.message.append(error_eq_0)

    def map_net_view(self):
        for network in self._networks.networks:
            net = network.network
            if net not in self.config.pyt:
                continue
            tmp = set()
            for view in self.config.pyt[net]:
                if (
                    self._is_site_included(network.site_name, view)
                    and self._is_vrf_included(network.vrf, view)
                    and self._is_net_allowed(net, self.config.view_dict[view])
                ):
                    tmp.add(view)
            self._check_views(
                network,
                tmp,
                f"Multiple matching Network Views based on Network Config found: {','.join(tmp)}",
                "No matching Network Views found based on Network Config will continue with VRF, Site, and Default.",
                "Matched Network based on Network Mapping.",
            )

    def _map_view(self, mapping, attr, error_gt_1: str, error_eq_0: str, success: str):
        for network in self._networks.networks:
            if network.net_view:
                continue
            elif attr == "vrf" and not getattr(network, attr):
                network.error.no_matching_view = True
                network.error.message.append("Network has no VRF assigned.")
                continue
            tmp = set()
            for view in mapping[getattr(network, attr)].include:
                if (
                    self._is_site_included(network.site_name, view.name)
                    and self._is_vrf_included(network.vrf, view.name)
                    and self._is_net_allowed(network.network, view)
                ):
                    tmp.add(view.name)
            self._check_views(network, tmp, f"{error_gt_1}{','.join(tmp)}", error_eq_0, success)

    def map_vrf_view(self):
        self._map_view(
            self._vrf_map,
            "vrf",
            "Multiple matching Network Views based on VRF Config found: ",
            "No matching Network Views found based on VRF Config will continue with Site and Default.",
            "Matched Network based on VRF Mapping.",
        )

    def map_site_view(self):
        self._map_view(
            self._site_map,
            "site_name",
            "Multiple matching Network Views based on Site Name Config found: ",
            "No matching Network Views found based on Site Name Config will continue with Default.",
            "Matched Network based on Site Name Mapping.",
        )

    def map_default_view(self):
        default = self.config.default_view
        for network in self._networks.networks:
            if network.net_view:
                continue
            if not self.config.default_view:
                network.error.no_matching_view = True
                network.error.message.append("Network did not match any rules and no default assigned.")
            tmp = set()
            if (
                self._is_site_included(network.site_name, default.name)
                and (network.vrf in ["", None] or self._is_vrf_included(network.vrf, default.name))
                and self._is_net_allowed(network.network, default)
            ):
                tmp.add(default)
                network.net_view = default
                network.error.multiple_views, network.error.no_matching_view = False, False
                network.success.network_view = "Matched Network based on Default Mapping."
            else:
                network.error.no_matching_view = True
                network.error.message.append("Network did not match any rules and excluded by default.")



from ipaddress import IPv4Interface, IPv4Network
from typing import Optional

from pydantic import BaseModel, Field, PrivateAttr, ConfigDict


class Success(BaseModel):
    network_view: Optional[str] = None


class Error(BaseModel):
    multiple_views: bool = False
    no_matching_view: bool = False
    message: Optional[list[str]] = Field(default_factory=list)


class ManagedNetwork(BaseModel):
    site_name: str = Field(alias="siteName")
    network: IPv4Interface = Field(alias="net")
    vrf: Optional[str] = None
    mapped_vrf: Optional[str] = None
    # vlan: Optional[int] = Field(alias="vlanId")
    _net_view: str = PrivateAttr(None)
    _error: Error = PrivateAttr(default_factory=Error)
    _success: Success = PrivateAttr(default_factory=Success)
    _skip_reason: str = PrivateAttr(None)
    model_config = ConfigDict(extra="ignore")

    def __hash__(self):
        return hash(f"{self.network} - {self.site_name} - {str(self.mapped_vrf)}")

    @property
    def error(self):
        return self._error

    @property
    def success(self):
        return self._success

    @property
    def net_view(self):
        return self._net_view

    @net_view.setter
    def net_view(self, value):
        self._net_view = value


class ManagedNetworks(BaseModel):
    networks: list[ManagedNetwork]

    @property
    def sites(self) -> set[str]:
        return {_.site_name for _ in self.networks}

    @property
    def vrfs(self) -> set[str]:
        return {_.vrf for _ in self.networks if _.vrf}

    @property
    def nets(self) -> list[IPv4Network]:
        return [_.network.network for _ in self.networks]

    @property
    def duplicate_nets(self) -> set[IPv4Network]:  # TODO
        return {_ for _ in self.nets if self.nets.count(_) > 1}



from typing import Optional, Union

from httpx import URL
from httpx._types import VerifyTypes
from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from .models import TimeoutTypes


class Infoblox(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", env_prefix="ib_", extra="ignore")
    host: Union[str, URL] = Field(None)
    verify_ssl: Optional[VerifyTypes] = True
    timeout: Union[TimeoutTypes] = 5.0
    username: str
    password: str

    @field_validator("verify_ssl")
    @classmethod
    def _verify(cls, v: Union[bool, int, str]) -> bool:
        if isinstance(v, (int, bool)):
            return bool(v)
        return False if v.lower() in {0, "0", "off", "f", "false", "n", "no"} else True
